#include <Arduino.h>
#include <NextionLightLib.h>
#include <mini-printf.h>

// un-comment the following line to get I/O debug information on the Serial monitor
// #define NXDEBUG 1

Nextion::Nextion(Stream& _serial, NextionType t, bool orientation, uint16_t bufsize) : myserial(_serial)
{
   maxx = 320;
   maxy = 240;
   extended = false;
   expect_result = false;
   rcvbufsize = bufsize;
   rcvbuf = new uint8_t[bufsize];
   switch(t)
   {
   case NXTYPE_NX32: maxx=400; maxy=240; break;
   case NXTYPE_NX35: maxx=480; maxy=320; break;
   case NXTYPE_NX43: maxx=480; maxy=272; break;
   case NXTYPE_NX50: 
   case NXTYPE_NX70: maxx=800; maxy=480; break;
   case NXTYPE_NX24E: 
   case NXTYPE_NX28E: extended = true; break;
   case NXTYPE_NX32E: maxx=400; maxy=240; extended = true; break;
   case NXTYPE_NX35E: maxx=480; maxy=320; extended = true; break;
   case NXTYPE_NX43E: maxx=480; maxy=272; extended = true; break;
   case NXTYPE_NX50E: 
   case NXTYPE_NX70E: maxx=800; maxy=480; extended = true; break;
   };
   if(orientation) // vertical, swap dimensions!
   {
      unsigned int w = maxx;
      maxx = maxy;
      maxy = w;
   }
}

Nextion::~Nextion()
{
   delete rcvbuf;
}

// begin() - initialize communication with Nextion
// NOTE: the serial interface has to be initialized (serialx.begin(9600)) 
//       BEFORE calling Nextion.begin()!
NextionError Nextion::begin(const char *startscreen)
{
   NextionError rc = NXSUCCESS;

   myserial.setTimeout(1000);		// wait max. 1s for returns

   sendCmd("");				// trigger baud detect on Nextion
   delay(100);				// wait for it to finish
   set_bkcmd(3);			// always request a return code
   expect_result = true;		// from now on we must get return data

   if(startscreen && *startscreen) // Inital screen given
   {
      unsigned int len = strlen(startscreen)+6;
      char cmd[len];
      mini_snprintf(cmd, len, "page %s", startscreen);
      rc = sendCmd(cmd);
   }
   return rc;
}


// sendCmd() - start an action on the Nextion and return the raw result data
NextionError Nextion::sendCmd(const char *cmd, unsigned int expected)
{
   NextionError rc = NXSUCCESS;
   unsigned char *cp = rcvbuf;
   unsigned int bytes_read = 0;

#ifdef NXDEBUG
   Serial.println(cmd);
#endif

   if(expected && expected>rcvbufsize) return NXERR_RECEIVE_BUFFER_OVERFLOW;

   while(myserial.available()) myserial.read(); // clean up myserial buffer
   // send command and end marker 0xFFFFFF
   myserial.print(cmd);
   myserial.write(0xFF);
   myserial.write(0xFF);
   myserial.write(0xFF);
   
   rc = expected ? getRaw(expected) : getTerminated(); 	// try to read the requested data
   return rc;
}

NextionError Nextion::getNum(const char *obj, unsigned int *val)
{
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "get %s", obj);
   rc = sendCmd(buf);
   if(rc==NXOK_NUMERIC_DATA)
   {
      *val = rcvbuf[1]+(rcvbuf[2]<<8)+(rcvbuf[3]<<16)+(rcvbuf[4]<<24);
   }
   return rc;
}

NextionError Nextion::setNum(const char *obj, unsigned int val)
{
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "%s=%u", obj, val);
   rc = sendCmd(buf);
   return rc;
}

NextionError Nextion::getAtt(const char *obj, const char *att, unsigned int *val)
{
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "get %s.%s", obj, att);
   rc = sendCmd(buf);
   if(rc==NXOK_NUMERIC_DATA)
   {
      *val = rcvbuf[1]+(rcvbuf[2]<<8)+(rcvbuf[3]<<16)+(rcvbuf[4]<<24);
   }
   return rc;
}

NextionError Nextion::setAtt(const char *obj, const char *att, unsigned int val)
{
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "%s.%s=%u", obj, att, val);
   rc = sendCmd(buf);
   return rc;
}

NextionError Nextion::get_txt(const char *obj, char *target, unsigned int maxlen)
{
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];
   unsigned char *cp = rcvbuf+1;

   mini_snprintf(buf, VALBUFSIZE, "get %s.txt", obj);
   rc = sendCmd(buf);
   if(rc==NXOK_STRING_DATA)
   {
      strncpy(target, (const char *)cp, maxlen);
      if(strlen((const char *)cp)>maxlen)
      {
	 rc = NXERR_TARGET_STRING_OVERFLOW;
	 target[maxlen-1] = 0;
      }
   }
   return rc;
}

NextionError Nextion::set_txt(const char *obj, const char *source)
{
   NextionError rc = NXSUCCESS;
   char buf[TXTBUFSIZE];

   mini_snprintf(buf, TXTBUFSIZE, "%s.txt=\"%s\"", obj, source);
   rc = sendCmd(buf);
   return rc;
}


NextionError Nextion::getEEPROMnum(int offset, uint32_t *val)
{
   if(!extended) return NXERR_INVALID_DEVICE;
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "rept %d,4", offset);
   rc = sendCmd(buf, 4);
   if(rc==NXOK_TRANSPARENT_DATA_FINISHED)
   {
      *val = rcvbuf[0]+(rcvbuf[1]<<8)+(rcvbuf[2]<<16)+(rcvbuf[3]<<24);
   }
   return rc;
}

NextionError Nextion::setEEPROMnum(int offset, uint32_t val)
{
   if(!extended) return NXERR_INVALID_DEVICE;
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "wept %d,4", offset);
   rc = sendCmd(buf);
   if(rc==NXOK_READY_FOR_TRANSPARENT_DATA)
   {
      uint8_t databyte;
      delay(25);
      for(uint8_t i=0;i<4;++i)
      {
	 databyte = (val & 0xFF);
	 myserial.write(databyte);
	 val >>= 8;
      }
   }
   rc = getTerminated();
   return rc;
}

NextionError Nextion::getEEPROMtxt(int offset, char *target, unsigned int maxlen)
{
   if(!extended) return NXERR_INVALID_DEVICE;
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "rept %d,%d", offset, maxlen);
   rc = sendCmd(buf, maxlen);
   if(rc==NXOK_TRANSPARENT_DATA_FINISHED)
   {
      memcpy(target, (const char *)rcvbuf, maxlen);
   }
   return rc;
}

NextionError Nextion::setEEPROMtxt(int offset, char *source, unsigned int len)
{
   if(!extended) return NXERR_INVALID_DEVICE;
   NextionError rc = NXSUCCESS;
   char buf[VALBUFSIZE];

   mini_snprintf(buf, VALBUFSIZE, "wept %d,%d", offset, len);
   rc = sendCmd(buf);
   if(rc==NXOK_READY_FOR_TRANSPARENT_DATA)
   {
      delay(25);
      for(uint8_t i=0;i<len;++i)
      {
	 myserial.write(source[i]);
      }
   }
   rc = getTerminated();
   return rc;
}

// getRaw() - get a predefined number of raw bytes from the Nextion via UART
NextionError Nextion::getRaw(unsigned int expected)
{
   NextionError rc = NXSUCCESS;
   unsigned char *cp = rcvbuf;
   long tout;
   bytes_read = 0;

#ifdef NXDEBUG
   Serial.println("getRaw()");
#endif
   if(expected>rcvbufsize) return NXERR_RECEIVE_BUFFER_OVERFLOW;

   tout = millis()+NEXTION_SERIAL_TIMEOUT;
   do
   {
      if(myserial.available())
      {
	 uint8_t c = myserial.read();
	 *cp++ = c;
	 bytes_read++;
	 tout = millis()+NEXTION_SERIAL_TIMEOUT;
      }
      delay(5); // give Nextion some more time to provide next byte
   } while(millis()<tout && bytes_read<expected);

   // Not all expected bytes read? Must be timeout.
   if(bytes_read<expected) rc=NXERR_RECEIVE_TIMEOUT;
   else                    rc=NXOK_TRANSPARENT_DATA_FINISHED;
#ifdef NXDEBUG
   mini_uprintf(Serial,"%03X - %016B\n", rc, rcvbuf);
#endif
   return rc;
}

// getTerminated() - get data from the Nextion via UART until the end marker 0xFFFFFF is found
NextionError Nextion::getTerminated(void)
{
   NextionError rc = NXSUCCESS;
   unsigned char *cp = rcvbuf;
   uint8_t FFcnt = 0;
   long tout;
   bytes_read = 0;

#ifdef NXDEBUG
   Serial.println("getTerminated()");
#endif
   tout = millis()+NEXTION_SERIAL_TIMEOUT;
   do
   {
      if(myserial.available())
      {
         uint8_t c = myserial.read();
         *cp++ = c;
         bytes_read++;
         if(0xFF==c) FFcnt++; else FFcnt = 0;	// end marker byte?
         tout = millis()+NEXTION_SERIAL_TIMEOUT;
      }
      delay(5); // give Nextion some more time to provide next byte
   } while(millis()<tout && FFcnt<3 && bytes_read<rcvbufsize);

   *cp = 0;
   // end marker not yet found and no buffer overflow? Must be timeout.
   if(FFcnt<3 && bytes_read<rcvbufsize) rc=NXERR_RECEIVE_TIMEOUT;
   else
   {
      if(bytes_read>=rcvbufsize) rc=NXERR_RECEIVE_BUFFER_OVERFLOW;
      else if(3==FFcnt)
      {
         *(cp-3) = 0;	// remove trailing end marker
	 bytes_read -= 3;
         rc = (NextionError)rcvbuf[0];
      }
      else 
      {
         rc = NXERR_RECEIVE_NOT_TERMINATED;
      }
   }
#ifdef NXDEBUG
   mini_uprintf(Serial,"%03X - %016B\n", rc, rcvbuf+1);
#endif
   return rc;
}

NextionError Nextion::getPage(unsigned int *val)
{
   NextionError rc = NXSUCCESS;
   unsigned char *cp = rcvbuf;

   rc = sendCmd("sendme");
   if(NXOK_PAGE_ID==rc)
   {
      *val = (unsigned int)(*(cp+1));
   }
   return rc;
}

NextionError Nextion::getTouch(unsigned int *x, unsigned int *y, bool *event)
{
   NextionError rc = NXSUCCESS;

   rc = getTerminated();
   if(NXOK_TOUCH_COORDINATES==rc || NXOK_TOUCH_EVENT_SLEEP==rc)
   {
      *x = rcvbuf[2]+(rcvbuf[1]<<8);
      *y = rcvbuf[4]+(rcvbuf[3]<<8);
      *event = rcvbuf[5] ? true : false;
   }
   return rc;
}

NextionError Nextion::getClick(unsigned int *page, unsigned int *component, bool *event)
{
   NextionError rc = NXSUCCESS;

   rc = getTerminated();
   if(NXOK_TOUCH_EVENT==rc)
   {
      *page = rcvbuf[1];
      *component = rcvbuf[2];
      *event = rcvbuf[3];
   }
   return rc;
}

NextionError Nextion::getDate(unsigned int *year, unsigned int *month, unsigned int *day, unsigned int *weekday)
{
   NextionError rc = NXSUCCESS;
   unsigned int val;

   if(!extended) return NXERR_INVALID_DEVICE;
   rc = get_rtc0(&val);
   if(rc==NXOK_NUMERIC_DATA)
   {
      *year = val;
      rc = get_rtc1(&val);
      if(rc==NXOK_NUMERIC_DATA)
      {
	 *month = val;
	 rc = get_rtc2(&val);
	 if(rc==NXOK_NUMERIC_DATA)
	 {
	    *day = val;
	    rc = get_rtc6(&val);
	    *weekday = val;
	 }
      }
   }
   return rc;
}

NextionError Nextion::getTime(unsigned int *hour, unsigned int *minute, unsigned int *second)
{
   NextionError rc = NXSUCCESS;
   unsigned int val;

   if(!extended) return NXERR_INVALID_DEVICE;
   rc = get_rtc3(&val);
   if(rc==NXOK_NUMERIC_DATA)
   {
      *hour = val;
      rc = get_rtc4(&val);
      if(rc==NXOK_NUMERIC_DATA)
      {
	 *minute = val;
	 rc = get_rtc5(&val);
	 *second = val;
      }
   }
   return rc;
}

NextionError Nextion::getRawBuffer(uint8_t *buffer, unsigned int *numbytes)
{
   NextionError rc = NXSUCCESS;

   if(!bytes_read) rc = NXERR_NO_RETURN_DATA;
   else if(bytes_read<(*numbytes))
   {
      memcpy(buffer, rcvbuf, bytes_read);
      *numbytes = bytes_read;
   }
   else rc = NXERR_TARGET_STRING_OVERFLOW;
   return rc;
}
