#ifndef _NEXTION_H
#define _NEXTION_H

// Some preprocessor magic to encode all set/get functions
// Functions will be named "set_att()" / "get_att()", "att" standing for all Nextion attributes
// like bco, font, maxlen etc.
// SETGETVAL(att) always includes GETVAL(att)
#define GETVAL(_att) inline NextionError get_##_att(const char *obj, unsigned int *val) { return getAtt(obj, (const char *)#_att, val); }
#define SETGETVAL(_att) GETVAL(_att) \
   inline NextionError set_##_att(const char *obj, unsigned int val) { return setAtt(obj, (const char *)#_att, val); }

// Same as for the attributes, applied here to all system variables
#define GETSYS(_att) inline NextionError get_##_att(unsigned int *val) { return getNum((const char *)#_att, val); }
#define SETGETSYS(_att) GETSYS(_att) \
   inline NextionError set_##_att(unsigned int val) { return setNum((const char *)#_att, val); }

// Two #defines for the internal commandstrings to get/set numeric and string values
// If you plan to use very long object names you may have to extend these buffer sizes
#define VALBUFSIZE 64
#define TXTBUFSIZE 128

// define the internal timeout to wait for another byte of data on UART
#define NEXTION_SERIAL_TIMEOUT 100

/*
Model 			Size 	Resolution 	Color 	Flash 	enum	
Nextion Basic Models
NX3224T024_011R 	2.4 	320*240 	65536 	4 	NX24	
NX3224T028_011R 	2.8 	320*240 	65536 	4 	NX28
NX4024T032_011R 	3.2 	400*240 	65536 	4 	NX32
NX4832T035_011R 	3.5 	480*320 	65536 	16 	NX35
NX4827T043_011R 	4.3 	480*272 	65536 	16 	NX43
NX8048T050_011R 	5.0 	800*480 	65536 	16	NX50
NX8048T070_011R 	7.0 	800*480 	65536 	16 	NX70
Nextion Enhanced Models
NX3224K024_011R 	2.4 	320*240 	65536 	16 	NX24E
NX3224K028_011R 	2.8 	320*240 	65536 	16 	NX28E
NX4024K032_011R 	3.2 	400*240 	65536 	16 	NX32E
NX4832K035_011R 	3.5 	480*320 	65536 	32 	NX35E
NX4827K043_011R 	4.3 	480*272 	65536 	32 	NX43E
NX8048K050_011R 	5.0 	800*480 	65536 	32 	NX50E
NX8048K070_011R 	7.0 	800*480 	65536 	32 	NX70E
*/

// Nextion model type codes to determine maximum pixel dimensions, capabilities etc.
typedef enum
{
   NXTYPE_NX24, NXTYPE_NX28, NXTYPE_NX32, NXTYPE_NX35, NXTYPE_NX43, NXTYPE_NX50, NXTYPE_NX70,
   NXTYPE_NX24E, NXTYPE_NX28E, NXTYPE_NX32E, NXTYPE_NX35E, NXTYPE_NX43E, NXTYPE_NX50E, NXTYPE_NX70E,
} NextionType;

// Error code list.
// Almost every function in this class returns such an error code. You better care about them... ;-)
// Codes below 0xFF are standard Nextion codes, some above 0x0400 have been added for internal
// results from the class functions
typedef enum
{
   NXERR_INVALID_INSTRUCTION = 0x00,
   NXSUCCESS = 0x01,
   NXERR_INVALID_COMPONENT_ID = 0x02,
   NXERR_INVALID_PAGE_ID = 0x03,
   NXERR_INVALID_PICTURE_ID = 0x04,
   NXERR_INVALID_FONT_ID = 0x05,
   NXERR_INVALID_BAUD_RATE = 0x11,
   NXERR_INVALID_CHANNEL_ID = 0x12,
   NXERR_INVALID_VARIABLE_NAME = 0x1A,
   NXERR_INVALID_VARIABLE_OPERATION = 0x1B,
   NXERR_ASSIGNMENT_FAILED = 0x1C,
   NXERR_PERFORM_FAILED = 0x1D,
   NXERR_INVALID_PARAMETER_COUNT = 0x1E,
   NXERR_IO_OPERATION_FAILED = 0x1F,
   NXERR_UNDEFINED_ESCAPE_CHARS = 0x20,
   NXERR_VARIABLE_NAME_TOO_LONG = 0x23,
   NXOK_TOUCH_EVENT = 0x65,
   NXOK_PAGE_ID = 0x66,
   NXOK_TOUCH_COORDINATES = 0x67,
   NXOK_TOUCH_EVENT_SLEEP = 0x68,
   NXOK_STRING_DATA = 0x70,
   NXOK_NUMERIC_DATA = 0x71,
   NXOK_AUTO_SLEEP = 0x86,
   NXOK_AUTO_WAKE = 0x87,
   NXOK_STARTUP = 0x88,
   NXOK_SD_UPGRADE = 0x89,
   NXOK_TRANSPARENT_DATA_FINISHED = 0xFD,
   NXOK_READY_FOR_TRANSPARENT_DATA = 0xFE,
   NXERR_INVALID_DEVICE = 0x0401,
   NXERR_RECEIVE_BUFFER_OVERFLOW = 0x0402,
   NXERR_RECEIVE_NOT_TERMINATED = 0x0403,
   NXERR_NO_RETURN_DATA = 0x0404,
   NXERR_RECEIVE_TIMEOUT = 0x0405,
   NXERR_TARGET_STRING_OVERFLOW = 0x0406,
} NextionError;

class Nextion
{
   protected:
      NextionType typ;			// Nextion model
      unsigned int maxx, maxy;		// width, height in pixels
      bool extended;			// true, if extended model with RTC
      Stream& myserial;			// Serial from/to Nextion
      uint8_t *rcvbuf;			// Buffer to hold bytes received from Nextion
      uint16_t rcvbufsize;		// Buffer size
      unsigned int bytes_read;		// Bytes read with last turn
      bool expect_result;		// switch to wait for return data from Nextion or not
      					// Only set to false during initialization
     
      // getNum - get a numeric value from Nextion
      NextionError getNum(const char *obj, unsigned int *val);
     
      // setNum - set a numeric value on the Nextion
      NextionError setNum(const char *obj, unsigned int val);
     
      // getAtt - get a numeric attribute for an object
      NextionError getAtt(const char *obj, const char *att, unsigned int *val);
     
      // setAtt - set a numeric attribute for an object
      NextionError setAtt(const char *obj, const char *att, unsigned int val);
     
      // getTerminated - read data from Nextion until the end marker 0xFFFFFF is encountered
      NextionError getTerminated(void);
     
      // getRaw - read a set number of bytes from the Nextion
      NextionError getRaw(unsigned int expected);
      
   public:
      // Constructor
      // _serial: reference to an initialized serial interface
      // orientation: false=landscape, true=upright
      // bufsize: number of bytes in the receive buffer for data from the Nextion
      Nextion(Stream& _serial, NextionType t, bool orientation=false, uint16_t bufsize=256);
     
      // Destructor
      ~Nextion();
     
      // begin - initialize the Nextion object
      // startscreen: objname of the page to visit initially
      NextionError begin(const char *startscreen);
     
      // sendCmd - send a command line to the NExtion to have it processed
      // expected: if 0, a terminated response (0xFFFFFF) is expected. (getTerminated)
      //           all other values will read the given number of bytes (getRaw)
      NextionError sendCmd(const char *cmd, unsigned int expected=0);
     
      // getTouch - get touch coordinates after "sendxy=1" was issued
      // x, y: pointers to the target variables for coordinates
      // event: pointer to target variable for touch event
      NextionError getTouch(unsigned int *x, unsigned int *y, bool *event);
     
      // getClick - get object IDs being touched (send ID must have been set in the Nextion editor)
      // page, component: pointers to target variables
      // event: pointer to target variable for touch event
      NextionError getClick(unsigned int *page, unsigned int *component, bool *event);

      // Object attribute helper functions
      // SETGETVAL means there is a set_xxx() and a get_xxx() function for attribute xxx
      // GETVAL    creates a get_xxx() function only for read-only attributes
      SETGETVAL(val)
      GETVAL(id)
      SETGETVAL(objname)
      GETVAL(vscope)
      GETVAL(mode)
      GETVAL(sta)
      GETVAL(psta)
      GETVAL(style)
      GETVAL(isbr)
      GETVAL(pw)
      SETGETVAL(bco)
      SETGETVAL(bco2)
      SETGETVAL(pco)
      SETGETVAL(pco2)
      SETGETVAL(pco0)
      SETGETVAL(font)
      SETGETVAL(xcen)
      SETGETVAL(ycen)
      SETGETVAL(spax)
      SETGETVAL(spay)
      SETGETVAL(x)
      SETGETVAL(y)
      SETGETVAL(w)
      SETGETVAL(h)
      SETGETVAL(dir)
      SETGETVAL(tim)
      SETGETVAL(dis)
      SETGETVAL(en)
      SETGETVAL(lenth)	// Chinese for 'length' :-)
      SETGETVAL(pic)
      SETGETVAL(picc)
      SETGETVAL(wid)
      SETGETVAL(hig)
      SETGETVAL(maxval)
      SETGETVAL(minval)
      SETGETVAL(ch)
      SETGETVAL(gdc)
      SETGETVAL(gdw)
      SETGETVAL(gdh)
    
      // System variable helper functions
      // SETGETSYS means there is a set_xxx() and a get_xxx() function for system variable xxx
      // GETSYS    creates a get_xxx() function only for read-only system variables
      SETGETSYS(baud)
      SETGETSYS(bauds)
      SETGETSYS(dim)
      SETGETSYS(dims)
      SETGETSYS(spax)
      SETGETSYS(spay)
      SETGETSYS(bkcmd)
      SETGETSYS(sendxy)
      SETGETSYS(thc)
      SETGETSYS(thdra)
      SETGETSYS(ussp)
      SETGETSYS(thsp)
      SETGETSYS(thup)
      SETGETSYS(delay)
      SETGETSYS(sleep)
      GETSYS(rand)
      SETGETSYS(sys0)
      SETGETSYS(sys1)
      SETGETSYS(sys2)
      SETGETSYS(rtc0)
      SETGETSYS(rtc1)
      SETGETSYS(rtc2)
      SETGETSYS(rtc3)
      SETGETSYS(rtc4)
      SETGETSYS(rtc5)
      GETSYS(rtc6)
      SETGETSYS(pio0)
      SETGETSYS(pio1)
      SETGETSYS(pio2)
      SETGETSYS(pio3)
      SETGETSYS(pio4)
      SETGETSYS(pio5)
      SETGETSYS(pio6)
      SETGETSYS(pio7)
      SETGETSYS(pwm4)
      SETGETSYS(pwm5)
      SETGETSYS(pwm6)
      SETGETSYS(pwm7)
      SETGETSYS(pwmf)
      SETGETSYS(wup)

      // get_txt - get the txt attribute of an object
      // target: the receiving char buffer
      // maxlen: the maximum number of chars tolerated
      NextionError get_txt(const char *obj, char *target, unsigned int maxlen);
     
      // set_txt - set the txt attribute of an object
      // source: the char buffer holding the chars to be set
      NextionError set_txt(const char *obj, const char *source);
     
      // Next two functions need to be coded explicitly, as the C preprocessor stumbles upon the "-" in "txt-maxl"...
      // get_txt_maxl and set_txt_maxl - NOTE the underscore!
      inline NextionError get_txt_maxl(const char *obj, unsigned int *val) { return getAtt(obj, (const char *)"txt-maxl", val); }
      inline NextionError set_txt_maxl(const char *obj, unsigned int val) { return setAtt(obj, (const char *)"txt-maxl", val); }
     
      // getPage - get the currently displayed page number
      // val: pointer to variable to put the page number into
      NextionError getPage(unsigned int *val);
     
      // rgb2color - convert R, G, B values into Nextion 565 color format
      inline unsigned int rgb2color(uint8_t red, uint8_t green, uint8_t blue) { return (((red>>3)&0x1F)<<11)|(((green>>2)&0x3F)<<5)|((blue>>3)&0x1F); }
     
      // getEEPROMnum - get numeric value from EEPROM
      // offset: address in EEPROM
      // val: 32bit variable to hold the returned number. NOTE: must be handled with "long" datatype on Arduino Uno!
      NextionError getEEPROMnum(int offset, uint32_t *val);
     
      // setEEPROMnum - store a 32bit number into EEPROM
      // offset: address in EEPROM
      // val: 32bit variable holding the number to be stored
      NextionError setEEPROMnum(int offset, uint32_t val);
     
      // getEEPROMtxt - get a sequence of bytes from EEPROM
      // offset: address in EEPROM
      // target: char buffer to hold the returned bytes
      // maxlen: maximum number of bytes target can hold
      NextionError getEEPROMtxt(int offset, char *target, unsigned int maxlen);
     
      // setEEPROMtxt - store a sequence of bytes in EEPROM
      // offset: address in EEPROM
      // source: buffer holding the byte sequence
      // len: number of bytes to be written
      NextionError setEEPROMtxt(int offset, char *source, unsigned int len);
     
      // getDate - get day, month, year and weekday from RTC
      // year, month, day, weekday: target variables to hold the returned data
      NextionError getDate(unsigned int *year, unsigned int *month, unsigned int *day, unsigned int *weekday);
     
      // getTime - get current time from RTC
      // hour, minute, second: target variables to hold the returned data
      NextionError getTime(unsigned int *hour, unsigned int *minute, unsigned int *second);
     
      // getRawBuffer - read the last data returned from Nextion. Nextion data is held in an internal buffer
      //                and will be copied to the supplied buffer
      // buffer: array to hold the copied data
      // numbytes: pointer to a variable holding the maximum number of bytes buffer can hold.
      //           upon return this variable is filled with the number of bytes effectively returned
      NextionError getRawBuffer(uint8_t *buffer, unsigned int *numbytes);
};
#endif
